#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "kbs_Alphabet.h"
#include "kbs_Error.h"
#include "kbs_Limits.h"
#include "kbs_Math.h"
#include "kbs_String.h"
#include "kbs_SuffixArray.h"
#include "kbs_SuffixArrayConstDStepAndPreNumbers.h"
#include "kbs_Time.h"
#include "kbs_Types.h"


__inline__ static void insSortPtrLevelNum_SaBucket(register Kbs_Long *const leftPtr, Kbs_Long *const rightPtr, register const Kbs_Ulong offset, register const Kbs_Long *const sufIndexMap);
__inline__ static Kbs_Long MedianOfThreeLong(const Kbs_Long a, const Kbs_Long b, const Kbs_Long c);
static void qSortInsSortPtrLevelNum_SaBucket(Kbs_Long *const leftPtr, Kbs_Long *const rightPtr, const Kbs_Ulong offset, const Kbs_Long *const sufIndexMap);

__inline__ static void computeBucketSize2Num_SaBucket(register const Kbs_Long leftIndex, register const Kbs_Long rightIndex, const Kbs_Ulong d, register const Kbs_Ulong offset, register Kbs_Long *const sufIndexMap, register Kbs_Ulong *const suffixArray);
static Kbs_Long computeDiffDepthBucketNum_SaBucket(register const Kbs_Long leftIndex, register const Kbs_Long rightIndex, const Kbs_Ulong d, register const Kbs_Ulong offset, register Kbs_Long *const sufIndexMap, register Kbs_Ulong *const suffixArray);

static void updatePtrAndRefineBucketsNum_SaBucket(const Kbs_Long leftIndex, const Kbs_Long rightIndex, const Kbs_Ulong d, Kbs_Ulong offset, Kbs_Long *sufIndexMap, Kbs_Ulong *const suffixArray);

static void insSortViaPtrNum_SaBucket(const Kbs_Long leftIndex, const Kbs_Long rightIndex, const Kbs_Ulong d, const Kbs_Ulong offset, Kbs_Long *sufIndexMap, Kbs_Ulong *const suffixArray);
static void qSortInsSortViaPtrNum_SaBucket(const Kbs_Long leftIndex, const Kbs_Long rightIndex, const Kbs_Ulong d, const Kbs_Ulong offset,Kbs_Long *sufIndexMap, Kbs_Ulong *const suffixArray);

static Kbs_Long *determineAll_Buckets_Sarray_Num_SuffixArray(const Kbs_Ustring *const mappedString, register const Kbs_Ulong d, Kbs_Ulong **suffixArrayPtr);
static Kbs_Long *determinePower2Alpha_Buckets_Sarray_Num_SuffixArray(const Kbs_Ustring *const mappedString, register const Kbs_Ulong d, Kbs_Ulong **suffixArrayPtr);
static Kbs_Long *determine_Buckets_Sarray_Num_SuffixArray(const Kbs_Ustring *const mappedString, register const Kbs_Ulong d, Kbs_Ulong **suffixArrayPtr);

static Kbs_Long *determineAll_Buckets_Sarray_Sindexmap_Num_SuffixArray(const Kbs_Ustring *const mappedString, register const Kbs_Ulong d, Kbs_Ulong **suffixArrayPtr, Kbs_Long **sufIndexMapPtr);
static Kbs_Long *determinePower2Alpha_Buckets_Sarray_Sindexmap_Num_SuffixArray(const Kbs_Ustring *const mappedString, register const Kbs_Ulong d, Kbs_Ulong **suffixArrayPtr, Kbs_Long **sufIndexMapPtr);
static Kbs_Long *determine_Buckets_Sarray_Sindexmap_Num_SuffixArray(const Kbs_Ustring *const mappedString, register const Kbs_Ulong d, Kbs_Ulong **suffixArrayPtr, Kbs_Long **sufIndexMapPtr);




/**
 * Determines the maximal q-gram size d for the given alphabet size and
 * @param alphabetSize
 * @return maximal q-gram size
 */
__inline__ static Kbs_Ulong determineMaxD(const Kbs_Ulong alphabetSize) {

	if (alphabetSize == 1) {
		return 10000;
	}
	Kbs_Ulong numberHashValues = 1;
	Kbs_Ulong d = 0;
	const Kbs_Ulong bound = KBS_ULONG_MAX / (alphabetSize * sizeof(Kbs_Long));
	while (numberHashValues <= bound){
		numberHashValues *= alphabetSize;
		d++;
	}
	return d;
}



/**
 * Performs the stable sort of one level of a bucket concerning the pointers in the sufPtrMap
 * @param leftPtr points to the leftmost position of the current bucket
 * @param rightPtr points to the rightmost position of the current bucket
 * @param offset denotes that contained suffixes share same prefix of length offset, which is used to refer to the sort key
 * @param sufIndexMap at position i is index of the rightmost position of the bucket in the suffix array that contains suffix i (bptr)
 */
__inline__ static void insSortPtrLevelNum_SaBucket(register Kbs_Long *const leftPtr,
																			Kbs_Long *const rightPtr,
																			register const Kbs_Ulong offset,
																			register const Kbs_Long *const sufIndexMap) {

	register Kbs_Long *rightTmpPtr = leftPtr+1;
	while(rightTmpPtr <= rightPtr) {
		register const Kbs_Long tempValue = *rightTmpPtr;
		register const Kbs_Long tempHashValue = sufIndexMap[(*rightTmpPtr)+offset];
		register Kbs_Long *leftTmpPtr = rightTmpPtr;
		while(leftTmpPtr > leftPtr && sufIndexMap[*(leftTmpPtr-1)+offset] > tempHashValue) {
			*leftTmpPtr = *(leftTmpPtr-1);
			leftTmpPtr--;
		}
		*leftTmpPtr = tempValue;
		rightTmpPtr++;
	}
}


/**
 * @param a first key
 * @param b second key
 * @param c third key
 * @return median of a,b,c
 */
__inline__ static Kbs_Long MedianOfThreeLong(const Kbs_Long a, const Kbs_Long b, const Kbs_Long c) {
	if (a==b) {
		return 0;
	}
	if (a==c || b==c) {
		return 2;
	}
	return a<b ?(b<c ? 1 : (a<c ? 2 : 0)) : (b>c ? 1 : (a<c ? 0 : 2));
}



/**
 * Mixture of Quicksort and Insertion Sort
 * Partitions smaller than a given threshold are sorted by insertion sort
 * @param leftPtr points to the leftmost position of the current bucket
 * @param rightPtr points to the rightmost position of the current bucket
 * @param offset denotes that contained suffixes share same prefix of length offset, which is used to refer to the sort key
 * @param sufIndexMap at position i is index of the rightmost position of the bucket in the suffix array that contains suffix i (bptr)
 */
static void qSortInsSortPtrLevelNum_SaBucket(Kbs_Long *const leftPtr,
																		register Kbs_Long *const rightPtr,
																		register const Kbs_Ulong offset,
																		const Kbs_Long *const sufIndexMap) {

	register const Kbs_Long halfBucketSize = (Kbs_Long)(rightPtr - leftPtr)/2;
	register Kbs_Long pivot = sufIndexMap[*rightPtr + offset];
	register const Kbs_Long pivotb = sufIndexMap[*leftPtr + offset];
	register const Kbs_Long pivotc = sufIndexMap[*(leftPtr+halfBucketSize) + offset];
	register const Kbs_Int medNumber = MedianOfThreeLong(pivot, pivotb, pivotc);
	if (medNumber > 0){
		register Kbs_Long *pivotPtr = (medNumber == 1) ? leftPtr : (leftPtr+halfBucketSize);
		pivot = (medNumber == 1) ? pivotb : pivotc;
		register const Kbs_Long swapTmp = *pivotPtr;
		*pivotPtr = *rightPtr;
		*rightPtr = swapTmp;
	}
	register Kbs_Long *leftTmpPtr = leftPtr;
	register Kbs_Long *rightTmpPtr = leftPtr;
	while (rightTmpPtr < rightPtr) {
		if (sufIndexMap[*rightTmpPtr + offset] <= pivot) {
			register const Kbs_Long swapTmp = *leftTmpPtr;
			*leftTmpPtr = *rightTmpPtr;
			*rightTmpPtr = swapTmp;
			leftTmpPtr++;
		}
		rightTmpPtr++;
	}
	register const Kbs_Long swapTmp = *leftTmpPtr;
	*leftTmpPtr = *rightTmpPtr;
	*rightTmpPtr = swapTmp;
	register Kbs_Long partSize;
	if ((partSize = rightPtr-leftTmpPtr) > 1) {
 		if (partSize < KBS_INSSORT_THRES_LEN) { /* sort by insertion sort */
 			insSortPtrLevelNum_SaBucket(leftTmpPtr+1, rightPtr, offset, sufIndexMap);
 		}
 		else {
			qSortInsSortPtrLevelNum_SaBucket(leftTmpPtr+1, rightPtr, offset, sufIndexMap);
		}
	}
	do {
		leftTmpPtr--;
	} while (leftTmpPtr > leftPtr && sufIndexMap[*leftTmpPtr+offset] == pivot);
	if ((partSize = leftTmpPtr-leftPtr) > 0) {
 		if (partSize < KBS_INSSORT_THRES_LEN) { /* sort by insertion sort */
 			insSortPtrLevelNum_SaBucket(leftPtr, leftTmpPtr, offset, sufIndexMap);
 		}
 		else {
			qSortInsSortPtrLevelNum_SaBucket(leftPtr, leftTmpPtr, offset, sufIndexMap);
 		}
	}
}


/**
 * Completely sort buckets of size 2
 * @param leftIndex points to the left suffix of the current bucket
 * @param rightIndex points to the right suffix of the current bucket
 * @param d is the initial prefix length used for the bucket sort. It also determines the increase of offset
 * @param offset denotes that contained suffixes share same prefix of length offset, which is used to refer to the sort key
 * @param sufIndexMap at position i is index of the rightmost position of the bucket in suffixArray that contains suffix i (bptr)
 * @param suffixArray array of suffixes to be sorted
 */
__inline__ static void computeBucketSize2Num_SaBucket(register const Kbs_Long leftIndex,
																		register const Kbs_Long rightIndex,
																		const Kbs_Ulong d,
																		register const Kbs_Ulong offset,
																		register Kbs_Long *const sufIndexMap,
																		register Kbs_Ulong *const suffixArray) {
	register Kbs_Long suffix1 = suffixArray[leftIndex] + offset;
	register Kbs_Long suffix2 = suffixArray[rightIndex] + offset;
	while (sufIndexMap[suffix1] == sufIndexMap[suffix2]) {
		suffix1 += d;
		suffix2 += d;
	}
	if (sufIndexMap[suffix1] > sufIndexMap[suffix2]) {
		register Kbs_Long tmpSwap = suffixArray[leftIndex];
		suffixArray[leftIndex] = suffixArray[rightIndex];
		suffixArray[rightIndex] = tmpSwap;
	}
	sufIndexMap[suffixArray[leftIndex]] = leftIndex;
}


/**
 * Computes the LCP of all suffixes in this bucket. This LCP will be the newoffset
 * @param leftIndex points to the left suffix of the current bucket
 * @param rightIndex points to the right suffix of the current bucket
 * @param d is the initial prefix length used for the bucket sort. It also determines the increase of offset
 * @param offset denotes that contained suffixes share same prefix of length offset, which is used to refer to the sort key
 * @param sufIndexMap at position i is index of the rightmost position of the bucket in suffixArray that contains suffix i (bptr)
 * @param suffixArray array of suffix numbers to be sorted
 */
static Kbs_Long computeDiffDepthBucketNum_SaBucket(register const Kbs_Long leftIndex,
																		register const Kbs_Long rightIndex,
																		const Kbs_Ulong d,
																		register const Kbs_Ulong offset,
																		register Kbs_Long *const sufIndexMap,
																		register Kbs_Ulong *const suffixArray) {
	register Kbs_Long lcp = offset;
	while (1) {
		register Kbs_Long i;
		for (i=leftIndex; i< rightIndex; i++) {
			if(sufIndexMap[suffixArray[i]+lcp] != sufIndexMap[suffixArray[i+1]+lcp]) {
				return lcp;
			}
		}
		lcp +=d;
	}
}

/**
 * The function determines the subbuckets of this bucket
 * and recursively calls the refinement function for the sorting of 2 levels in one step from left to right
 * @param leftIndex index of the suffix array indicating the leftmost index of the current bucket
 * @param rightIndex index of the suffix array indicating the rightmost index of the current bucket
 * @param d is the initial prefix length used for the bucket sort. It also determines the increase of offset
 * @param offset denotes that contained suffixes share the same prefix of length offset, which is used to refer to the sort key
 * @param sufIndexMap is index of the rightmost position of the bucket that contains suffix i
 * @param suffixArray array of suffix numbers to be sorted
 */
static void updatePtrAndRefineBucketsNum_SaBucket(register const Kbs_Long leftIndex,
																		register const Kbs_Long rightIndex,
																		const Kbs_Ulong d,
																		register const Kbs_Ulong offset,
																		register Kbs_Long *const sufIndexMap,
																		register Kbs_Ulong *const suffixArray) {

	/* for all buckets with resp. pointer > rightIndex determine buckets via setting sufIndexMap */
	register Kbs_Long leftIntervalIndex = rightIndex;
	register Kbs_Long rightIntervalIndex = rightIndex;
	register Kbs_Long tmpIndex;
	while(leftIndex <= leftIntervalIndex && rightIndex < (tmpIndex = sufIndexMap[suffixArray[leftIntervalIndex] + offset] )) {
		do{
			sufIndexMap[suffixArray[leftIntervalIndex]] = rightIntervalIndex;
			leftIntervalIndex--;
		} while (leftIndex <= leftIntervalIndex
							&& sufIndexMap[suffixArray[leftIntervalIndex] + offset] == tmpIndex);
		rightIntervalIndex = leftIntervalIndex;
	}

	/* since the sufIndexMap for the suffixes between leftIndex and rightIndex might change in previous 2 steps */
	/* determine the bucket concerning suffixptr+offset between leftIndex and rightIndex seperately */
	rightIntervalIndex = leftIntervalIndex;
	while (leftIndex <= leftIntervalIndex
					&& leftIndex <= sufIndexMap[suffixArray[leftIntervalIndex]+offset]
					&& sufIndexMap[suffixArray[leftIntervalIndex]+offset] <= rightIndex) {
		sufIndexMap[suffixArray[leftIntervalIndex]] = rightIntervalIndex;
		leftIntervalIndex--;
	}
	const Kbs_Long middleRightIndex = rightIntervalIndex;
	const Kbs_Long middleLeftIndex = leftIntervalIndex;

	/* for all buckets with resp. pointer+offset < leftIndex determine buckets via setting sufIndexMap */
	/* start with rightIntervalIndex which indicates leftend-1 of bucket with resp. pointer+offset between*/
	/* leftIndex and rightIndex*/
	rightIntervalIndex = leftIntervalIndex;
	while(leftIndex <= leftIntervalIndex) {
		register const Kbs_Long tmpInd2 = sufIndexMap[suffixArray[leftIntervalIndex] + offset];
		do{
			sufIndexMap[suffixArray[leftIntervalIndex]] = rightIntervalIndex;
			leftIntervalIndex--;
		} while (leftIndex <= leftIntervalIndex
							&& sufIndexMap[suffixArray[leftIntervalIndex] + offset] == tmpInd2);
		rightIntervalIndex = leftIntervalIndex;
	}

	Kbs_Long newOffset = offset+d;
	if (sufIndexMap[suffixArray[leftIndex]]==rightIndex) {
		newOffset = computeDiffDepthBucketNum_SaBucket(leftIndex, rightIndex, d, newOffset, sufIndexMap, suffixArray);
	}
	register Kbs_Long leftTmpIndex = leftIndex;
	while (leftTmpIndex < middleLeftIndex) {
		const Kbs_Long rightTmpIndex = sufIndexMap[suffixArray[leftTmpIndex]];
		const Kbs_Long tmpLong = rightTmpIndex - leftTmpIndex;
		if (tmpLong > 0) {
			if (tmpLong == 1) {
				computeBucketSize2Num_SaBucket(leftTmpIndex, rightTmpIndex, d, newOffset, sufIndexMap, suffixArray);
				leftTmpIndex = rightTmpIndex + 1;
				continue;
			}
			if (tmpLong < KBS_INSSORT_THRES_LEN) { /* sort by insertion sort */
 				insSortViaPtrNum_SaBucket(leftTmpIndex, rightTmpIndex, d, newOffset, sufIndexMap, suffixArray);
 			}
 			else
				qSortInsSortViaPtrNum_SaBucket(leftTmpIndex, rightTmpIndex, d, newOffset, sufIndexMap, suffixArray);
 		}
		leftTmpIndex = rightTmpIndex + 1;
	}
	/* for buckets refering to this bucket, the offset can be doubled */
	if (middleRightIndex > middleLeftIndex+1) {
		if (middleRightIndex - middleLeftIndex == 2) {
			computeBucketSize2Num_SaBucket(middleLeftIndex+1, middleRightIndex, d, 2*offset, sufIndexMap, suffixArray);
		}
		else {
			if (middleRightIndex - middleLeftIndex < KBS_INSSORT_THRES_LEN) { /* sort by insertion sort */
 				insSortViaPtrNum_SaBucket(middleLeftIndex+1, middleRightIndex, d, 2*offset, sufIndexMap, suffixArray);
 			}
 			else {
				qSortInsSortViaPtrNum_SaBucket(middleLeftIndex+1, middleRightIndex, d, 2*offset, sufIndexMap, suffixArray);
			}
		}
	}

	leftTmpIndex = middleRightIndex+1;
	while (leftTmpIndex < rightIndex) {
		const Kbs_Long rightTmpIndex = sufIndexMap[suffixArray[leftTmpIndex]];
		const Kbs_Long tmpLong = rightTmpIndex - leftTmpIndex;
		if (tmpLong > 0) {
			if (tmpLong == 1) {
				computeBucketSize2Num_SaBucket(leftTmpIndex, rightTmpIndex, d, newOffset, sufIndexMap, suffixArray);
				leftTmpIndex = rightTmpIndex + 1;
				continue;
			}
			if (tmpLong < KBS_INSSORT_THRES_LEN) { /* sort by insertion sort */
 				insSortViaPtrNum_SaBucket(leftTmpIndex, rightTmpIndex, d, newOffset, sufIndexMap, suffixArray);
 			}
 			else
				qSortInsSortViaPtrNum_SaBucket(leftTmpIndex, rightTmpIndex, d, newOffset, sufIndexMap, suffixArray);
 		}
		leftTmpIndex = rightTmpIndex + 1;
	}
}


/**
 * Refine this bucket by using insertion sort internally
 * @param leftIndex index of the suffix array indicating the leftmost index of the current bucket
 * @param rightIndex index of the suffix array indicating the rightmost index of the current bucket
 * @param d is the initial prefix length used for the bucket sort. It also determines the increase of offset
 * @param offset denotes that contained suffixes share the same prefix of length offset, which is used to refer to the sort key
 * @param sufIndexMap is index of the rightmost position of the bucket that contains suffix i
 * @param suffixArray array of suffix numbers to be sorted
*/
static void insSortViaPtrNum_SaBucket(Kbs_Long leftIndex,
																		Kbs_Long rightIndex,
																		const Kbs_Ulong d,
																		const Kbs_Ulong offset,
																		Kbs_Long *sufIndexMap,
																		Kbs_Ulong *const suffixArray) {
	/* use stable sort for the bucket (insertion sort) */
	insSortPtrLevelNum_SaBucket(suffixArray + leftIndex, suffixArray + rightIndex, offset, sufIndexMap);
	updatePtrAndRefineBucketsNum_SaBucket(leftIndex, rightIndex, d, offset, sufIndexMap, suffixArray);
}




/**
 * Refine this bucket by using mixed Quick-Insertion-Sort internally
 * @param leftIndex index of the suffix array indicating the leftmost index of the current bucket
 * @param rightIndex index of the suffix array indicating the rightmost index of the current bucket
 * @param d is the initial prefix length used for the bucket sort. It also determines the increase of offset
 * @param offset denotes that contained suffixes share the same prefix of length offset, which is used to refer to the sort key
 * @param sufIndexMap is index of the rightmost position of the bucket that contains suffix i
 * @param suffixArray array of suffix numbers to be sorted
*/
static void qSortInsSortViaPtrNum_SaBucket(Kbs_Long leftIndex,
																		Kbs_Long rightIndex,
																		const Kbs_Ulong d,
																		const Kbs_Ulong offset,
																		Kbs_Long *sufIndexMap,
																		Kbs_Ulong *const suffixArray) {
	qSortInsSortPtrLevelNum_SaBucket(suffixArray + leftIndex, suffixArray + rightIndex, offset, sufIndexMap);
	updatePtrAndRefineBucketsNum_SaBucket(leftIndex, rightIndex, d, offset, sufIndexMap, suffixArray);
}






/**
 * Constructs all q-gram buckets and the underlying up to prefix d sorted suffix array
 * Don't call this function directly,
 * use determine_Buckets_Sarray_Num_SuffixArray instead.
 * @param mappedString   - String of characters 0..alphabetsize-1
 * @param d              - size of q-gram
 * @param suffixArrayPtr - ptr that will be filled with the suffix array
 * @return buckets that contains ptr's into the suffix array
 * @see determinePower2Alpha_Buckets_Sarray_Num_SuffixArray
 * @see determine_Buckets_Sarray_Num_SuffixArray
 */
static Kbs_Long *determineAll_Buckets_Sarray_Num_SuffixArray(const Kbs_Ustring *const mappedString,
																					register const Kbs_Ulong d,
																					Kbs_Ulong **suffixArrayPtr) {

	register const Kbs_Ulong alphabetSize = mappedString->alphabet->alphaSize;
	register const Kbs_Long numberBuckets = kbs_power_Ulong(alphabetSize, d);
	register Kbs_Long *const buckets = (Kbs_Long *)malloc(sizeof(Kbs_Long) * (numberBuckets+1));
	memset(buckets, 0, sizeof(Kbs_Long) * (numberBuckets+1));

	/* computation of first hashvalue */
	register Kbs_Uchar *mappedUcharArray = mappedString->str;
	register Kbs_Ulong hashCode = 0;
	Kbs_Ulong tempPower = 1;
	register Kbs_Long j;
	for (j=d-1; j>=0; j--) {
		hashCode += mappedUcharArray[j] * tempPower;
		tempPower *= alphabetSize;
	}
	const Kbs_Ulong firstHashCode = hashCode;

	/* computation of the filling of buckets */
	register const Kbs_Long strLen = mappedString->strLength;
	mappedUcharArray += d;
	register const Kbs_Ulong tempModulo = kbs_power_Ulong(alphabetSize, d-1);
	for (j=0; j<strLen; j++) {
		buckets[hashCode]++;
		hashCode = hashCode % tempModulo;
		hashCode *= alphabetSize;
		hashCode += *mappedUcharArray;
		mappedUcharArray++;
	}

	/* computation of the bucket pointers, pointers into the suffix array */
	for (j=1; j<=numberBuckets; j++){
		buckets[j] = buckets[j-1] + buckets[j];
	}

	/* computation of the suffix array */
	Kbs_Ulong *const suffixArray = (Kbs_Ulong *)malloc(sizeof(Kbs_Ulong) * strLen);
	if (suffixArray == NULL) {
		KBS_ERROR(KBS_ERROR_MALLOC);
	}
	mappedUcharArray = mappedString->str + d;
	hashCode = firstHashCode ;
	for (j=0; j<strLen; j++) {
		buckets[hashCode]--;
		suffixArray[buckets[hashCode]] = j;
		hashCode = hashCode % tempModulo;
		hashCode *= alphabetSize;
		hashCode += *mappedUcharArray;
		mappedUcharArray++;
	}
	buckets[numberBuckets] = strLen;
	*suffixArrayPtr = suffixArray;
	return buckets;
}

/**
 * Constructs all q-gram buckets and the underlying up to prefix d sorted suffix array
 * for alphabet of size alphabetsize=2^x
 * Don't call this function directly,
 * use determine_Buckets_Sarray_Num_SuffixArray instead.
 * @param mappedString   - String of characters 0..alphabetsize-1
 * @param d              - size of q-gram
 * @param suffixArrayPtr - ptr that will be filled with the suffix array
 * @return buckets that contains ptr's into the suffix array
 * @see determineAll_Buckets_Sarray_Num_SuffixArray
 * @see determine_Buckets_Sarray_Num_SuffixArray
 */
static Kbs_Long *determinePower2Alpha_Buckets_Sarray_Num_SuffixArray(const Kbs_Ustring *const mappedString,
																					register const Kbs_Ulong d,
																					Kbs_Ulong **suffixArrayPtr) {

	register const Kbs_Int exp2 = (Kbs_Int) kbs_getExp_Ulong(2, mappedString->alphabet->alphaSize);
	if (exp2 < 0) {
		KBS_ERROR(KBS_ERROR_VALUEOUTOFBOUNDS);
		return NULL;
	}
	/* initialization of bucket table */
	register const Kbs_Ulong numberBuckets = kbs_power_Ulong(mappedString->alphabet->alphaSize, d);
	register Kbs_Long *const buckets = (Kbs_Long *)malloc(sizeof(Kbs_Long) * (numberBuckets+1));
	memset(buckets, '\0', sizeof(Kbs_Long) * (numberBuckets+1));

	/* computation of first hashvalue */
	register Kbs_Uchar *mappedUcharArray = mappedString->str;
	register Kbs_Ulong j;
	register Kbs_Ulong hashCode = (Kbs_Ulong)0;
	for (j=0; j<d; j++) {
		hashCode = hashCode << exp2;
		hashCode += mappedUcharArray[j];
	}
	const Kbs_Ulong firstHashCode = hashCode;

	/* computation of the filling for the buckets */
	register Kbs_Ulong tempModulo = (Kbs_Long)0;
	tempModulo = ~tempModulo;
	tempModulo = tempModulo << (exp2*(d-1));
	tempModulo = ~tempModulo;
	register Kbs_Ulong strLen = mappedString->strLength;
	mappedUcharArray += d;
	buckets[hashCode]++;
	for (j=1; j<strLen; j++) {
		hashCode = hashCode & tempModulo;
		hashCode = hashCode << exp2;
		hashCode = hashCode | (Kbs_Long)(*mappedUcharArray);
		mappedUcharArray++;
		buckets[hashCode]++;
	}

	/* computation of the bucket pointers, pointers into the suffix array */
	for (j=1; j<=numberBuckets; j++){
		buckets[j] = buckets[j-1] + buckets[j];
	}

	/* computation of the suffix array */
	Kbs_Ulong *const suffixArray = (Kbs_Ulong *)malloc(sizeof(Kbs_Ulong) * strLen);
	if (suffixArray == NULL) {
		KBS_ERROR(KBS_ERROR_MALLOC);
	}
	mappedUcharArray = mappedString->str + d;
	hashCode = firstHashCode ;
	for (j=0; j<strLen; j++) {
		buckets[hashCode]--;
		suffixArray[buckets[hashCode]] = j;
		hashCode = hashCode & tempModulo;
		hashCode = hashCode << exp2;
		hashCode = hashCode | (Kbs_Long)(*mappedUcharArray);
		mappedUcharArray++;
	}
	buckets[numberBuckets] = strLen;
	*suffixArrayPtr = suffixArray;
	return buckets;
}



/**
 * Constructs all q-gram buckets and the underlying up to prefix d sorted suffix array
 * @param mappedString   - String of characters 0..alphabetsize-1
 * @param d              - size of q-gram
 * @param suffixArrayPtr - ptr that will be filled with the suffix array
 * @return buckets that contains ptr's into the suffix array
 * @see determineAll_Buckets_Sarray_Num_SuffixArray
 * @see determinePower2Alpha_Buckets_Sarray_Num_SuffixArray
 */
static Kbs_Long *determine_Buckets_Sarray_Num_SuffixArray(const Kbs_Ustring *const mappedString,
																					register const Kbs_Ulong d,
																					Kbs_Ulong **suffixArrayPtr) {
	if (kbs_getExp_Ulong(2, mappedString->alphabet->alphaSize) >= 0) {
		return determinePower2Alpha_Buckets_Sarray_Num_SuffixArray(mappedString, d, suffixArrayPtr);
	}
	else {
		return determineAll_Buckets_Sarray_Num_SuffixArray(mappedString, d, suffixArrayPtr);
	}
}



/**
 * Constructs all q-gram buckets and the underlying up to prefix d sorted suffix array
 * and the initial suffix index map.
 * Don't call this function directly,
 * use determine_Buckets_Sarray_Sindexmap_Num_SuffixArray instead.
 * @param mappedString   - String of characters 0..alphabetsize-1
 * @param d              - size of q-gram
 * @param suffixArrayPtr - ptr that will be filled with the suffix array
 * @param sufIndexMapPtr - will be filled with pointer to sufIndexMap table
 * @return buckets that contains ptr's into the suffix array
 * @see determinePower2Alpha_Buckets_Sarray_Sindexmap_Num_SuffixArray
 * @see determine_Buckets_Sarray_Sindexmap_Num_SuffixArray
 */
static Kbs_Long *determineAll_Buckets_Sarray_Sindexmap_Num_SuffixArray(const Kbs_Ustring *const mappedString,
																					register const Kbs_Ulong d,
																					Kbs_Ulong **suffixArrayPtr,
																					Kbs_Long **sufIndexMapPtr) {


	register Kbs_Long *const buckets = determineAll_Buckets_Sarray_Num_SuffixArray(mappedString, d, suffixArrayPtr);

	register Kbs_Long strLen = mappedString->strLength;
	register Kbs_Long *const sufIndexMap = (Kbs_Long *)malloc(sizeof(Kbs_Long) * (strLen+2*d+1));
	if (sufIndexMap == NULL) {
		KBS_ERROR(KBS_ERROR_MALLOC);
	}


	/* computation of first hashvalue */
	register Kbs_Long j;
	register const Kbs_Long alphabetSize = mappedString->alphabet->alphaSize;
	register Kbs_Uchar *mappedUcharArray = mappedString->str;
	register Kbs_Long tempPower = 1;
	register Kbs_Long hashCode = 0;
	for (j=d-1; j>=0; j--) {
		hashCode += mappedUcharArray[j] * tempPower;
		tempPower *= alphabetSize;
	}
	register Kbs_Long tempModulo = kbs_power_Ulong(alphabetSize, d-1);
	mappedUcharArray += d;
	for (j=0; j<strLen-(Kbs_Long)d; j++) {
		sufIndexMap[j] = (buckets[hashCode+1]) - 1;
		hashCode = hashCode % tempModulo;
		hashCode *= alphabetSize;
		hashCode += *mappedUcharArray;
		mappedUcharArray++;
	}
	const Kbs_Ulong *const suffixArray = *suffixArrayPtr;
	const Kbs_Long firstHashCode = hashCode;
	for (j = strLen-d; j<strLen; j++) {
		sufIndexMap[suffixArray[buckets[hashCode]]] = buckets[hashCode];
		(buckets[hashCode])++;
		hashCode = hashCode % tempModulo;
		hashCode *= alphabetSize;
		hashCode += *mappedUcharArray;
		mappedUcharArray++;
	}
	hashCode = firstHashCode;
	mappedUcharArray -= d;
	for (j = strLen-d; j<strLen; j++) {
		buckets[hashCode]--;
		hashCode = hashCode % tempModulo;
		hashCode *= alphabetSize;
		hashCode += *mappedUcharArray;
		mappedUcharArray++;
	}
	/* set the values in sufIndexMap[strLen..strLen+2*d] to [-1, -2, ..., -2*d]*/
	Kbs_Long i =(-1);
	for (j=strLen; j<=strLen+2*(Kbs_Long)d; j++) {
		sufIndexMap[j] = i--;
	}
	*sufIndexMapPtr = sufIndexMap;
	return buckets;
}



/**
 * Constructs all q-gram buckets and the underlying up to prefix d sorted suffix array
 * and the initial suffix index map
 * for alphabet of size alphabetsize=2^x.
 * Don't call this function directly,
 * use determine_Buckets_Sarray_Sindexmap_Num_SuffixArray instead.
 * @param mappedString   - String of characters 0..alphabetsize-1
 * @param d              - size of q-gram
 * @param suffixArrayPtr - ptr that will be filled with the suffix array
 * @param sufIndexMapPtr - will be filled with pointer to sufIndexMap table
 * @return buckets that contains ptr's into the suffix array
 * @see determineAll_Buckets_Sarray_Sindexmap_Num_SuffixArray
 * @see determine_Buckets_Sarray_Sindexmap_Num_SuffixArray
 */
static Kbs_Long *determinePower2Alpha_Buckets_Sarray_Sindexmap_Num_SuffixArray(const Kbs_Ustring *const mappedString,
																					register const Kbs_Ulong d,
																					Kbs_Ulong **suffixArrayPtr,
																					Kbs_Long **sufIndexMapPtr) {

	register Kbs_Ulong strLen = mappedString->strLength;
	const Kbs_Int exp2 = (Kbs_Int) kbs_getExp_Ulong(2, mappedString->alphabet->alphaSize);
	if (exp2 < 0) {
		KBS_ERROR(KBS_ERROR_VALUEOUTOFBOUNDS);
		return NULL;
	}

	/* initialization of bucket table */
	register Kbs_Long *const buckets = determinePower2Alpha_Buckets_Sarray_Num_SuffixArray(mappedString, d, suffixArrayPtr);
	register Kbs_Long *const sufIndexMap = (Kbs_Long *)malloc(sizeof(Kbs_Long) * (strLen+2*d+1));
	if (sufIndexMap == NULL) {
		KBS_ERROR(KBS_ERROR_MALLOC);
	}

	/* computation of first hashvalue */
	register Kbs_Uchar *mappedUcharArray = mappedString->str;
	register Kbs_Ulong hashCode = 0;
	register Kbs_Ulong j;
	for (j=0; j<d; j++) {
		hashCode = hashCode << exp2;
		hashCode += mappedUcharArray[j];
	}
	register Kbs_Ulong tempModulo = (Kbs_Ulong)0;
	tempModulo = ~tempModulo;
	tempModulo = tempModulo << (exp2*(d-1));
	tempModulo = ~tempModulo;
	mappedUcharArray += d;
	for (j=0; j<strLen-d; j++) {
		sufIndexMap[j] = (buckets[hashCode+1]) - 1;
		hashCode = hashCode & tempModulo;
		hashCode = hashCode << exp2;
		hashCode = hashCode | (Kbs_Long)(*mappedUcharArray);
		mappedUcharArray++;
	}
	Kbs_Ulong firstHashCode = hashCode;
	const Kbs_Ulong *const suffixArray = *suffixArrayPtr;
	for (j = strLen-d; j<strLen; j++) { /* set sufIndexMap for outranging suffixes which already reach final positions */
		sufIndexMap[suffixArray[buckets[hashCode]]] = buckets[hashCode];
		(buckets[hashCode])++;
		hashCode = hashCode & tempModulo;
		hashCode = hashCode << exp2;
		hashCode = hashCode | (Kbs_Long)(*mappedUcharArray);
		mappedUcharArray++;
	}
	hashCode = firstHashCode;
	mappedUcharArray -= d;
	for (j = strLen-d; j<strLen; j++) { /* reset the bucket pointers */
		buckets[hashCode]--;
		hashCode = hashCode & tempModulo;
		hashCode = hashCode << exp2;
		hashCode = hashCode | (Kbs_Long)(*mappedUcharArray);
		mappedUcharArray++;
	}
	Kbs_Long i =(-1);
	for (j=strLen; j<=strLen+2*d; j++) {
		sufIndexMap[j] = i--;
	}
	*sufIndexMapPtr = sufIndexMap;
	return buckets;
}



/**
 * Constructs all q-gram buckets and the underlying up to prefix d sorted suffix array
 * and the initial suffix index map.
 * @param mappedString   - String of characters 0..alphabetsize-1
 * @param d              - size of q-gram
 * @param suffixArrayPtr - ptr that will be filled with the suffix array
 * @param sufIndexMapPtr - will be filled with pointer to sufIndexMap table
 * @return buckets that contains ptr's into the suffix array
 * @see determinePower2Alpha_Buckets_Sarray_Sindexmap_Num_SuffixArray
 * @see determineAll_Buckets_Sarray_Sindexmap_Num_SuffixArray
 */
static Kbs_Long *determine_Buckets_Sarray_Sindexmap_Num_SuffixArray(const Kbs_Ustring *const mappedString,
																					register const Kbs_Ulong d,
																					Kbs_Ulong **suffixArrayPtr,
																					Kbs_Long **sufIndexMapPtr) {
	if (kbs_getExp_Ulong(2, mappedString->alphabet->alphaSize) >= 0) {
		return determinePower2Alpha_Buckets_Sarray_Sindexmap_Num_SuffixArray(mappedString, d, suffixArrayPtr, sufIndexMapPtr);
	}
	else {
		return determineAll_Buckets_Sarray_Sindexmap_Num_SuffixArray(mappedString, d, suffixArrayPtr, sufIndexMapPtr);
	}
}



/*---------------------------------------------------------------------------------------*/
/**
 * Sorts the alphabet concerning some weight concerning cc bucket size and alphabet frequency
 * Only works for mapped string with alphabet [0,alphaSize]
 */
static Kbs_Uchar *getCharWeightedOrder_Alphabet(const Kbs_Alphabet *const alphabet,
																											const Kbs_Long *const buckets,
																											const Kbs_Long bucketsInLevel2Bucket,
																											const Kbs_Long weightOnCC){

	register const Kbs_Int alphabetSize = alphabet->alphaSize;
	Kbs_Long charWeight[alphabetSize];
	register const Kbs_Long tmpBucketFactor = bucketsInLevel2Bucket * (alphabetSize + 1);
	register Kbs_Long i;
	for (i=0; i<alphabetSize; i++) {
		charWeight[i] = (Kbs_Long)alphabet->charFreq[i];
		charWeight[i] -= weightOnCC * (buckets[i * tmpBucketFactor + bucketsInLevel2Bucket ]-buckets[i * tmpBucketFactor]);
	}

	register Kbs_Uchar *targetCharArray = (Kbs_Uchar *) malloc((alphabetSize + 1) * sizeof(Kbs_Uchar));
	for (i=0; i < alphabetSize; i++) {
		targetCharArray[i] = (Kbs_Uchar)i;
	}
	for (i=1; i < (Kbs_Long)alphabet->alphaSize; i++) {
		register const Kbs_Long tmpWeight = charWeight[i];
		register Kbs_Long j = i;
		while(j>0 && tmpWeight < charWeight[targetCharArray[j-1]])  {
			targetCharArray[j] = targetCharArray[j-1];
			j--;
		}
		targetCharArray[j] = (Kbs_Uchar)i;
	}
	return targetCharArray;
}

/*---------------------------------------------------------------------------------------*/
Kbs_SuffixArray *kbs_buildDstepUsePrePlusCopyFreqOrderNum_SuffixArray(Kbs_Ustring *const seq, register const Kbs_Ulong d) {

	if (seq == NULL) {
		KBS_ERROR(KBS_ERROR_NULLPOINTER);
		return NULL;
	}
	if (seq->alphabet == NULL) {
		KBS_ERROR(KBS_ERROR_NULLPOINTER);
		return NULL;
	}
	if (d > determineMaxD(seq->alphabet->alphaSize) ){
		printf("q-gram size = %lu, results in to many buckets\n", d);
		KBS_ERROR(KBS_ERROR_VALUEOUTOFBOUNDS);
		return NULL;
	}
	if (seq->alphabet == NULL) {
		KBS_ERROR(KBS_ERROR_NULLPOINTER);
		return NULL;
	}
	kbs_sort_Alphabet(seq->alphabet);
	/* computation of bucket table */
	kbs_convert_Ustring2RangeExtend(seq, d+d);

	Kbs_Ulong *suffixArray;
	Kbs_Long *sufIndexMap;
	Kbs_Long *buckets = determine_Buckets_Sarray_Sindexmap_Num_SuffixArray(seq, d, &suffixArray, &sufIndexMap);
	if (buckets == NULL) {
		KBS_ERROR(KBS_ERROR_NULLPOINTER);
		return NULL;
	}

	/* Sorting of all buckets */
	register const Kbs_Uchar *const mappedCharPtr = seq->str;
	const Kbs_Long alphabetSize = seq->alphabet->alphaSize;
	const Kbs_Long bucketsInLevel2Bucket = kbs_power_Ulong(alphabetSize, d-2);
	const Kbs_Long bucketsInLevel1Bucket = bucketsInLevel2Bucket * alphabetSize;
	const Kbs_Uchar *const alphaOrder = getCharWeightedOrder_Alphabet(seq->alphabet, buckets, bucketsInLevel2Bucket, 1);
	Kbs_Uchar *const isNotSortedLevel1Char = (Kbs_Uchar *) malloc(sizeof(Kbs_Uchar) * alphabetSize);
	memset(isNotSortedLevel1Char, '\1', sizeof(Kbs_Uchar) * alphabetSize);

	/* Sort all level-1 buckets */
	Kbs_Long leftIndexList[alphabetSize];		/* just for the copy phase maintains all pointers to the right of the small buckets */
	Kbs_Long rightIndexList[alphabetSize];		/* just for the copy phase maintains all pointers to the right of the small buckets */
	Kbs_Long i;
	for (i=0; i < alphabetSize; i++) {
		register const Kbs_Long c1 = alphaOrder[i];
		/* sort buckets cd to cz */
		register Kbs_Long j;
		for (j=i+1; j < alphabetSize; j++) {
			register const Kbs_Long c2 = alphaOrder[j];
			register const Kbs_Long tmpUlong = c1 * bucketsInLevel1Bucket + c2 * bucketsInLevel2Bucket;
			register Kbs_Long k;
			for (k = tmpUlong; k < tmpUlong + bucketsInLevel2Bucket; k++) {
				register const Kbs_Long leftIndex = buckets[k];
				register const Kbs_Long rightIndex = buckets[k+1] - 1;
				register Kbs_Long bucketSize;
				if ((bucketSize = rightIndex - leftIndex) > 0) {
					if (bucketSize < KBS_INSSORT_THRES_LEN)
						insSortViaPtrNum_SaBucket(leftIndex, (Kbs_Long)rightIndex, d, d, sufIndexMap, suffixArray);
					else
						qSortInsSortViaPtrNum_SaBucket(leftIndex, (Kbs_Long)rightIndex, d, d, sufIndexMap, suffixArray);
				}
			}
		}

		/* copy left buckets of cx */
		for (j=0; j<alphabetSize; j++) {
			leftIndexList[j] = buckets[j * bucketsInLevel1Bucket + c1 * bucketsInLevel2Bucket];
		}
		register Kbs_Long leftIndex = buckets[c1 * bucketsInLevel1Bucket];
		while(leftIndex < leftIndexList[c1]) {
			register Kbs_Long tmpUlong;
			register Kbs_Uchar tmpUchar;
			if ( (tmpUlong = suffixArray[leftIndex]) && isNotSortedLevel1Char[tmpUchar = mappedCharPtr[--tmpUlong]]) {
				register Kbs_Long tmpUlongIndex;
				sufIndexMap[tmpUlong] = tmpUlongIndex = leftIndexList[tmpUchar];
				leftIndexList[tmpUchar] = tmpUlongIndex + 1;
				suffixArray[tmpUlongIndex] = tmpUlong;
			}
			leftIndex++;
		}

		/* copy right buckets of cx */
		for (j=0; j<alphabetSize; j++) {
			rightIndexList[j] = buckets[j * bucketsInLevel1Bucket + (c1+1) * bucketsInLevel2Bucket];
		}
		register Kbs_Long rightIndex = (buckets[(c1+1)*bucketsInLevel1Bucket]);
		while((Kbs_Long)leftIndex < rightIndex) {
			rightIndex--;
			register Kbs_Long tmpUlong;
			register Kbs_Uchar tmpUchar;
			if ((tmpUlong = suffixArray[rightIndex]) && isNotSortedLevel1Char[tmpUchar = mappedCharPtr[--tmpUlong]]) {
				register Kbs_Long tmpUlongIndex;
				sufIndexMap[tmpUlong] = rightIndexList[tmpUchar] = tmpUlongIndex = rightIndexList[tmpUchar] - 1;
				suffixArray[tmpUlongIndex] = tmpUlong;
			}
		}

		isNotSortedLevel1Char[c1] = 0;
	}
/* End Sorting of all buckets */
	free(alphaOrder);
	free(isNotSortedLevel1Char);
 	Kbs_SuffixArray *const outputSA = (Kbs_SuffixArray *)malloc(sizeof(Kbs_SuffixArray));
 	outputSA->str = seq;
 	outputSA->posArray = suffixArray;
 	outputSA->invPosArray = sufIndexMap;
	outputSA->bucketIndices = buckets;
	outputSA->qGramSize = d;
	return outputSA;
}


